# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import testflows._core.contrib.schema as config
from testflows._core.test import Module, Suite, Test, Step, NullStep
from testflows._core.test import (
    TestStep,
    TestCase,
    TestSuite,
    TestModule,
    TestBackground,
    TestOutline,
    TestSketch,
)
from testflows._core.test import Context, SharedContext
from testflows._core.test import (
    Feature,
    Background,
    Scenario,
    Example,
    Outline,
    Sketch,
    Combination,
)
from testflows._core.test import (
    Critical,
    Major,
    Minor,
    Check,
    Given,
    When,
    Then,
    And,
    But,
    By,
    Finally,
    Cleanup,
)
from testflows._core.test import TestFeature, TestScenario, TestCheck
from testflows._core.test import (
    loads,
    ordered,
    retry,
    retries,
    repeat,
    repeats,
    define,
    choose,
    either,
)
from testflows._core.has import has
from testflows._core.flags import Flags
from testflows._core.objects import (
    OK,
    XOK,
    Fail,
    XFail,
    Skip,
    Error,
    XError,
    Null,
    XNull,
)
from testflows._core.objects import (
    Name,
    Description,
    Uid,
    Tags,
    Args,
    Setup,
    Parallel,
    Executor,
)
from testflows._core.objects import (
    XFails,
    XFlags,
    XArgs,
    Repeats,
    Repeat,
    Retries,
    Retry,
    Onlys,
    Skips,
)
from testflows._core.objects import OnlyTags, SkipTags, Maps
from testflows._core.objects import (
    FFails,
    Skipped,
    Failed,
    XFailed,
    XErrored,
    Okayed,
    XOkayed,
)
from testflows._core.objects import (
    Attributes,
    Requirements,
    Specifications,
    Examples,
    ArgumentParser,
    Timeouts,
)
from testflows._core.objects import (
    Node,
    Tag,
    Argument,
    Attribute,
    Requirement,
    Specification,
    Metric,
    Value,
    Ticket,
    Timeout,
)
from testflows._core.objects import Secret, RSASecret
from testflows._core.baseobject import Table
from testflows._core.filters import The
from testflows._core.funcs import load, append_path, cleanup
from testflows._core.funcs import main, args, private_key
from testflows._core.funcs import metric, ticket, value, note, debug, trace, text
from testflows._core.funcs import attribute, requirement, tag
from testflows._core.funcs import input, current_time
from testflows._core.funcs import message, exception, ok, fail, skip, err
from testflows._core.funcs import (
    result,
    null,
    xok,
    xfail,
    xerr,
    xnull,
    pause,
    getsattr,
    always,
    Timer,
    timer,
)
from testflows._core.funcs import (
    current_dir,
    current_module,
    load_module,
    load_submodules,
)
from testflows._core.funcs import aslice, chunks
from testflows._core.flags import TE, UT, SKIP, EOK, EFAIL, EERROR, ESKIP
from testflows._core.flags import XOK, XFAIL, XERROR, XNULL
from testflows._core.flags import FAIL_NOT_COUNTED, ERROR_NOT_COUNTED, NULL_NOT_COUNTED
from testflows._core.flags import (
    PAUSE,
    PAUSE_BEFORE,
    PAUSE_AFTER,
    REPORT,
    DOCUMENT,
    MANUAL,
    AUTO,
)
from testflows._core.flags import MANDATORY, CLEAR, NOT_REPEATABLE
from testflows._core.flags import EANY, ERESULT, XRESULT
from testflows._core.flags import PARALLEL, NO_PARALLEL
from testflows._core import __author__, __version__, __license__
from testflows._core.parallel import join, top, current, previous
from testflows._core.parallel.executor.thread import (
    ThreadPoolExecutor as Pool,
    ThreadPoolExecutor as ThreadPool,
    SharedThreadPoolExecutor as SharedThreadPool,
)
from testflows._core.parallel.executor.asyncio import (
    AsyncPoolExecutor as AsyncPool,
    SharedAsyncPoolExecutor as SharedAsyncPool,
)
from testflows._core.parallel.executor.process import (
    ProcessPoolExecutor as ProcessPool,
    SharedProcessPoolExecutor as SharedProcessPool,
    process_service,
)

import testflows.core.parallel as parallel
import testflows.core.objects as objects
import testflows.core.name as name

import testflows._core.utils as utils
import testflows._core.contrib.rsa as rsa



# Copyright 2023 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import math

from collections import namedtuple
from itertools import product, islice, combinations


class CoveringArrayError(Exception):
    """Covering array error."""

    def __init__(self, combination, values):
        self.combination = combination
        self.values = values

    def __str__(self):
        return f"missing combination={self.combination},values={self.values}"


class X:
    """Don't care value."""

    pass


BestTest = namedtuple("BestTest", "test coverage bitmap")
Π = namedtuple("Π", "combinations bitmap")


class CoveringArray:
    """Generate covering array of specified `strength`
    for the given `parameters` where parameters
    is a dict[str, list[value]] where key is the parameter name,
    value is a list of values for a given parameter.

    Uses an algorithm described in the following paper

    2007, "IPOG: A General Strategy for T-Way Software Testing" by
    Yu Lei1, Raghu Kacker, D. Richard Kuhn, Vadim Okun, and James Lawrence

    :param parameters: parameters (dict[str, list[values]])
    :param strength: (optional) strength, default: 2
    """

    def __init__(self, parameters, strength=2, array=None):
        self.parameters = parameters
        self.strength = strength
        self.array = array if array is not None else self.generate()

    def __iter__(self):
        return iter(self.array)

    def prepare(self):
        """Returns parameters set and its map.

        Parameters set is a list[list] where
        first level is indexed keys and second level
        are indexed values for a given parameter.

        Parameters map can be used to convert keys and values
        indexes to actual parameters name and values.

        :param parameters: parameters dictionary dict[str, list[values]])
        """
        keys = list(self.parameters.keys())

        parameters_set = []
        parameters_map = {}

        for key in keys:
            v = self.parameters[key]
            # remove any duplicates from v while preserving consistent ordering
            v = [x for i, x in enumerate(v) if x not in v[:i]]
            parameters_map[key] = v
            parameters_set.append(list(range(len(v))))

        return parameters_set, parameters_map

    def combination_index(self, t, N, combination):
        """Return an index for a given combination in π.
        Combination is represented by a tuple containing
        parameters referenced by their indexes.

        https://stackoverflow.com/questions/37764429/algorithm-for-combination-index-in-array

        f(5, 3, [2,3,4]) = binom(5,3) - binom(5-2,3) - binom(5-3,2) - binom(5-4,1) =
                         = 10 - 1 - 1 - 1 = 6

        :param combination: tuple(parameter index,...)
        """

        t -= 1
        index = math.comb(N, t) - 1
        for i, p in enumerate(combination[:-1]):
            p += 1
            index -= math.comb(N - p, t - i)

        return index

    def combination_values_bitmap_index(self, combination, values):
        """Return an bit index in the bitmap for a given
        parameter combination and the specific combination of values
        for these parameters.

        Example:
        ```
        >>> for j,i in enumerate(list(itertools.product([0,1,2,3],[0,1,2],[0,1,2]))):
        ...    print(i, 9*i[0] + 3*i[1] + 1*i[2], j)
        [0,1,2,3],[0,1,2],[0,1,2]
        i         j       k
        Jv * Kv = 9
        Kv = 3
        1
        ```

        :param combination: combination of parameters
        :param values: specific combination of values for the parameters
        """
        value_lengths = [len(self.parameters_set[p]) for p in combination]

        index = 0
        for i, value in enumerate(values):
            index += math.prod(value_lengths[i + 1 :]) * value
        return index

    def bitmap_index_combination_values(self, index, combination):
        """Return values for a given bitmap index and
        combination of parameters.

        Example:
        ```
        (3, 1, 1) 15 15
        5  2  2

        4*i[0] + 2*i[1] + 1*i[2] = 15
        = 15//4 = 3 (15-12) = 3
        = 3//2 = 1 (3 - 2) = 1
        = 1//1 = 1
        = (3,1,1)
        ```

        :param index: bitmap index
        :param combination: combination of parameters
        """
        value_lengths = [len(self.parameters_set[p]) for p in combination]

        value = [0] * len(combination)

        c = index
        for i in range(len(combination)):
            j = math.prod(value_lengths[i + 1 :])
            value[i] = c // j
            c = c - value[i] * j

        return value

    def construct_π(self, i, t):
        """Construct a set t-way combinations of values
        involving parameter Pi and t-1 parameters among
        the first i-1 parameters.

        Return pi which is a dictionary of bitmaps where each bit
        represent if parameter value combination is covered (0) or not (1).

        Initially, the bitmap is set to all 1's.
        """
        # create all t-way combinations of Pi with P0 to Pi-1 columns
        t_way_combinations = [c + (i,) for c in combinations(range(i), t - 1)]

        # FIXME: value_lengths should be stored
        # FIXME: value coefficients should be stored
        π = Π(
            t_way_combinations,
            [
                (1 << math.prod([len(self.parameters_set[p]) for p in combination])) - 1
                for combination in t_way_combinations
            ],
        )
        return π

    def horizontal_extension(self, t, i, π):
        """Horizontal extension for parameter Pi."""

        # for each test τ = (v1, v2, …, vi-1) in tests
        for τ in range(len(self.tests)):
            test = self.tests[τ]
            # choose a value vi of Pi and replace τ with τ’ = (v1, v2, …, vi-1, vi) so that τ’ covers the
            # most number of combinations of values in π
            best = None
            # FIXME: if bitmap is all empty set value to don't care and don't try to pick any value

            for value in self.parameters_set[i]:
                new_test = test + [value]
                coverage, bitmap = self.calculate_coverage(t, i, new_test, π)
                if best is None or coverage >= best.coverage:
                    best = BestTest(new_test, coverage, bitmap)
            self.tests[τ] = best.test
            π = Π(π.combinations, best.bitmap)

        return π

    def vertical_extension(self, t, i, π):
        """Vertical extension for parameter Pi"""

        # for each combination of σ in π
        for combination in π.combinations:
            index = self.combination_index(t, i, combination)

            bitmap = π.bitmap[index]
            bitmap_index = -1

            if bitmap == 0:
                continue

            while bitmap:
                bitmap_index += 1
                if not bitmap & 1:
                    bitmap >>= 1
                    continue
                bitmap >>= 1

                values = self.bitmap_index_combination_values(bitmap_index, combination)

                covered = False

                for test in self.tests:
                    test_values = self.combination_values(test, combination)
                    if values == test_values:
                        # already covered
                        covered = True
                        break

                    # not covered, so either change an existing test, if possible
                    change_existing = True
                    for value, test_value in zip(values, test_values):
                        if value == test_value:
                            continue
                        elif test_value is X:
                            continue
                        else:
                            change_existing = False

                    if change_existing:
                        self.set_combination_values(test, values, combination)
                        covered = True
                        break

                if covered:
                    # move to the next combination of values
                    continue

                # or otherwise add a new test to cover σ and remove it from π
                new_test = [X] * len(self.parameters_set[: i + 1])
                self.set_combination_values(new_test, values, combination)
                self.tests.append(new_test)

        # replace all X with the first value for a given parameter
        for test in self.tests:
            for i, p in enumerate(test):
                if p is not X:
                    continue
                test[i] = self.parameters_set[i][0]

    def convert_tests_to_covering_array(self):
        """Convert tests to covering array.

        Covering array is a list of tests where each test
        is a dictionary of parameter names to values list[dict(name, value)]
        where tests exhaustively cover a chosen level of t-way combinations of
        parameters.

        Example:
        Parameters map:  {'a': [1, 2], 'b': ['b', 'd', 'c', 'a'], 'c': [10]}

        """
        ca = []
        parameter_names = list(self.parameters_map.keys())

        for test in self.tests:
            ca_test = {}
            for i, p in enumerate(test):
                parameter_name = parameter_names[i]
                ca_test[parameter_name] = self.parameters_map[parameter_name][p]
            ca.append(ca_test)

        return ca

    def combination_values(self, test, combination):
        """Return values of the combination covered by a given test."""
        values = [0] * len(combination)

        for i, parameter in enumerate(combination):
            values[i] = test[parameter]

        return values

    def set_combination_values(self, test, values, combination):
        """Set test values for a given combination."""
        for i, parameter in enumerate(combination):
            test[parameter] = values[i]

    def calculate_coverage(self, t, i, test, π):
        """Calculate coverage of the test for a given π combinations."""
        coverage = 0

        new_bitmap = list(π.bitmap)

        for combination in π.combinations:
            index = self.combination_index(t, i, combination)
            bitmap = π.bitmap[index]

            current_coverage = (bitmap).bit_count()

            values = self.combination_values(test, combination)

            bitmap_index = self.combination_values_bitmap_index(combination, values)

            bitmap = bitmap & ~(1 << bitmap_index)
            new_coverage = (bitmap).bit_count()
            coverage += current_coverage - new_coverage

            new_bitmap[index] = bitmap

        return coverage - current_coverage, new_bitmap

    def generate(self):
        """Generate covering array."""
        # t-strength is > 1 and <= number of parameters
        t = max(1, min(self.strength, len(self.parameters)))

        # convert parameters dictionary into a parameters set
        # which uses only indexes for parameter names and values
        self.parameters_set, self.parameters_map = self.prepare()

        # construct first tests using all possible combinations of values
        # for the first t-strength parameters
        self.tests = list(product(*self.parameters_set[:t]))

        for i in range(len(self.tests)):
            self.tests[i] = list(self.tests[i])

        for i in range(t, len(self.parameters_set)):
            π = self.construct_π(i=i, t=t)
            π = self.horizontal_extension(t, i, π)
            self.vertical_extension(t, i, π)

        return self.convert_tests_to_covering_array()

    def check(self):
        """Returns True if covering array covers all t-strength
        combination of the parameters or raises an error."""
        t = self.strength

        if not self.array:
            raise ValueError("covering array is empty")

        parameter_names = list(self.parameters.keys())

        for combination in combinations(parameter_names, self.strength):
            for values in product(
                *[self.parameters[parameter] for parameter in combination]
            ):
                covered = True
                for test in self.array:
                    covered = True
                    for i, parameter in enumerate(combination):
                        if test[parameter] != values[i]:
                            covered = False
                            break
                    if covered:
                        break
                if not covered:
                    raise CoveringArrayError(combination, values)

        return True

    def __str__(self):
        """Dump covering array representation to string."""
        lines = [f"CoveringArray({self.parameters},{self.strength})["]
        lines.append(f"{len(self.array)}")
        header = " ".join(str(v) for v in self.array[0].keys())

        lines.append(header)
        lines.append("-" * len(header))

        for test in self.array:
            lines.append(" ".join(str(v) for v in test.values()))
        lines.append("]")

        return "\n".join(lines)



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import testflows.settings as settings

from testflows._core.flags import Flags, SKIP, RETRY
from testflows._core.testtype import TestType
from testflows._core.message import Message
from testflows._core.cli.colors import color
from testflows._core.utils.timefuncs import strftimedelta

indent = " " * 2


def color_result(result, attrs=None):
    if attrs is None:
        attrs = ["bold"]
    if result.startswith("X"):
        return functools.partial(color, color="blue", attrs=attrs)
    elif result == "OK":
        return functools.partial(color, color="green", attrs=attrs)
    elif result == "Skip":
        return functools.partial(color, color="cyan", attrs=attrs)
    elif result == "Error":
        return functools.partial(color, color="yellow", attrs=attrs)
    elif result == "Fail":
        return functools.partial(color, color="red", attrs=attrs)
    elif result == "Null":
        return functools.partial(color, color="magenta", attrs=attrs)
    else:
        raise ValueError(f"unknown result {result}")


def add_result(msg, results):
    flags = Flags(msg["test_flags"])
    cflags = Flags(msg["test_cflags"])
    result = msg["result_type"]

    if getattr(TestType, msg["test_type"]) < TestType.Iteration:
        if not result.startswith("X"):
            return
    if flags & SKIP and settings.show_skipped is False:
        return
    if cflags & RETRY:
        return

    if not result in ("OK", "Skip"):
        results[msg["test_id"]] = (msg, result)


processors = {
    Message.RESULT.name: (add_result,),
}


def generate(results, divider, only_new=False):
    """Generate report"""
    if not results:
        return

    retries = ""
    xfails = ""
    fails = ""
    first_fails = []

    for entry in results:
        msg, result = results[entry]
        _color = color_result(result)

        out = _color("\u2718") + f" [ {_color(result)} ] '{msg['result_test']}'"
        if msg["result_reason"]:
            out += color(f" \u1405 {msg['result_reason']}", "white", attrs=["dim"])
        out += " " + color(
            "(" + strftimedelta(msg["message_rtime"]) + ")", "white", attrs=["dim"]
        )
        out += "\n"

        if result.startswith("X"):
            if not only_new:
                xfails += out
        else:
            if len(first_fails) < 1:
                first_fails.append(msg["result_test"])
            fails += out

    if retries:
        retries = color(f"{divider}Retries\n\n", "white", attrs=["bold"]) + retries

    if xfails:
        if not divider and retries:
            divider = "\n"
        xfails = color(f"{divider}Known\n\n", "white", attrs=["bold"]) + xfails

    if fails:
        if not divider and (xfails or retries):
            divider = "\n"
        fails = color(f"{divider}Failing\n\n", "white", attrs=["bold"]) + fails

        fails += color(
            f"\nDebugging\n\n",
            "white",
            attrs=["bold"],
        )
        fails += (
            color(
                "Rerun the first failing test by executing your test program with the '--only' option.",
                "white",
                attrs=["dim"],
            )
            + "\n"
        )
        for first_fail in first_fails:
            fails += (
                color("--only ", "white")
                + color(f"'{first_fail}/*'", "red", attrs=["bold"])
                + "\n"
            )

        fails += (
            color(
                "\nSee first failed test messages using the 'tfs show messages' command.",
                "white",
                attrs=["dim"],
            )
            + "\n"
        )
        for first_fail in first_fails:
            fails += (
                color("tfs show messages --log test.log ", "white")
                + color(f"'{first_fail}'", "red", attrs=["bold"])
                + "\n"
            )

    report = f"{retries}{xfails}{fails}"

    return report or None


def transform(stop, divider="\n", only_new=False):
    """Transform parsed log line into a short format.

    :param stop: stop event
    :param divider: report divider, default: `\n`
    :param only_new: output only new fails, default: `False`
    """
    line = None
    results = {}
    while True:
        if line is not None:
            processor = processors.get(line["message_keyword"], None)
            if processor:
                processor[0](line, results, *processor[1:])
            line = None

        if stop.is_set():
            line = generate(results, divider, only_new)

        line = yield line



# Copyright 2020 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import textwrap
import functools

import testflows.settings as settings

from testflows._core.flags import Flags, SKIP, RETRY, LAST_RETRY
from testflows._core.testtype import TestType
from testflows._core.message import Message
from testflows._core.name import split, sep, parentname
from testflows._core.utils.timefuncs import strftimedelta
from testflows._core.cli.colors import color
from .brisk import transform as brisk_transform
from .nice import transform as nice_transform
from .plain import transform as plain_transform

indent = " " * 2


def color_other(other):
    return color(other, "white", attrs=["dim"])


def color_keyword(keyword):
    return color(split(keyword)[-1], "white", attrs=["dim"])


def color_secondary_keyword(keyword):
    return color(split(keyword)[-1], "white", attrs=["bold", "dim"])


def color_test_name(name):
    return color(name, "white", attrs=[])


def color_result(result, icon=None):
    def result_icon(result):
        r = result.lstrip("X")
        if r == "OK":
            return "\u2714"
        elif r == "Skip":
            return "\u2704"
        elif r == "Error":
            return "\u2718"
        elif r == "Fail":
            return "\u2718"
        else:
            return "\u2718"

    if icon is None:
        icon = result_icon(result)
    if result.startswith("X"):
        return color(icon, "blue", attrs=["bold"])
    elif result == "OK":
        return color(icon, "green", attrs=["bold"])
    elif result == "Skip":
        return color(icon, "white", attrs=["dim"])
    elif result == "Error":
        return color(icon, "yellow", attrs=["bold"])
    elif result == "Fail":
        return color(icon, "red", attrs=["bold"])
    # Null
    return color(icon, "cyan", attrs=["bold"])


def get_type(msg):
    return getattr(TestType, msg["test_type"])


def format_prompt(
    msg,
    dump_transform=None,
    buffer=None,
    skip_for_dump=None,
    skip_for_buffer=None,
    only_new=None,
):
    lines = (msg["message"] or "").splitlines()
    icon = "\u270d  "
    if msg["message"].startswith("Paused"):
        icon = "\u270b "
    out = color(icon + lines[0], "yellow", attrs=["bold"])
    if len(lines) > 1:
        out += "\n" + color("\n".join(lines[1:]), "white", attrs=["dim"])
    return out


def format_input(
    msg,
    dump_transform=None,
    buffer=None,
    skip_for_dump=None,
    skip_for_buffer=None,
    only_new=None,
):
    print("input", msg["message"])
    out = color(msg["message"], "white")
    return out


def format_multiline(text, indent):
    first, rest = (text.rstrip() + "\n").split("\n", 1)
    first = first.strip()
    if first:
        first += "\n"
    out = f"{first}{textwrap.dedent(rest.rstrip())}".rstrip()
    out = textwrap.indent(out, indent + "  ")
    return out


def collect_dump_messages(buffer, skip_for_dump, test_id):
    """Collect messages to dump for the failing test."""

    msgs = []
    if test_id not in buffer or test_id in skip_for_dump[0]:
        return msgs

    msgs += buffer[test_id]
    buffer[test_id] = []

    # get all child messages
    for name in buffer.keys():
        if name.startswith(test_id + sep):
            _msgs = buffer[name]
            if not _msgs:
                continue
            msgs += _msgs
            buffer[name] = []

    skip_test = test_id
    while skip_test and skip_test != sep:
        skip_for_dump[0].add(skip_test)
        skip_test = parentname(skip_test)

    msgs.sort(key=lambda x: x["message_time"])

    return msgs


def format_test(msg, dump_transform, buffer, skip_for_dump, skip_for_buffer, only_new):
    flags = Flags(msg["test_flags"])
    test_id = msg["test_id"]

    if flags & RETRY and not flags & LAST_RETRY:
        skip_for_buffer[0].add(test_id)

    return None


def format_result(
    msg, dump_transform, buffer, skip_for_dump, skip_for_buffer, only_new
):
    result = msg["result_type"]

    flags = Flags(msg["test_flags"])
    cflags = Flags(msg["test_cflags"])
    test_id = msg["test_id"]

    if flags & SKIP and settings.show_skipped is False:
        return

    if getattr(TestType, msg["test_type"]) < TestType.Iteration:
        return

    if cflags & RETRY and not flags & LAST_RETRY:
        return

    _icon = color_result(result)
    _result = color_result(result, result)
    _test = color_test_name(f"{msg['result_test']}")

    out = (
        f"{_icon} "
        + color_other(f"{strftimedelta(msg['message_rtime']):<10}")
        + f"[ {result.center(6, ' ')} ]".ljust(10, " ").replace(result, _result)
    )

    _result_message = msg["result_message"]
    if _result_message and settings.trim_results and int(msg["test_level"]) > 1:
        _result_message = _result_message.strip().split("\n", 1)[0].strip()

    if result in ("Fail", "Error", "Null"):
        out += f" {_test}"
        _test_msgs = collect_dump_messages(buffer, skip_for_dump, test_id)
        if _test_msgs:
            g = dump_transform(show_input=True)
            g.send(None)
            out += "\n"
            for _test_msg in _test_msgs:
                g_out = g.send(_test_msg)
                if g_out is not None:
                    out += g_out
            out = out.rstrip()
        if _result_message:
            out += f"\n{indent}  {color(format_multiline(_result_message, indent).lstrip(), 'yellow', attrs=['bold'])}"
        out += "\n"
    elif not only_new and result.startswith("X"):
        out += f" {_test}"
        if msg["result_reason"]:
            out += f"\n{indent}  {color(msg['result_reason'], 'blue', attrs=['bold'])}"
        out += "\n"
    else:
        out = None

    skip_for_buffer[0].discard(test_id)
    return out


formatters = {
    Message.INPUT.name: (format_input,),
    Message.PROMPT.name: (format_prompt,),
    Message.TEST.name: (format_test,),
    Message.RESULT.name: (format_result,),
}


def transform(
    brisk=False, plain=False, nice=False, pnice=False, only_new=False, show_input=True
):
    """Transform parsed log line into a fails format.

    :param dump: dump messages of the failing test
    :param only_new: output only new fails
    """
    line = None
    buffer = {}
    skip_for_dump = [set()]
    skip_for_buffer = [set()]

    dump_transform = None
    if brisk:
        dump_transform = brisk_transform
    if plain:
        dump_transform = plain_transform
    if nice:
        dump_transform = nice_transform
    if pnice:
        dump_transform = functools.partial(nice_transform, add_test_name_prefix=True)

    while True:
        skip = False
        if line is not None:
            if dump_transform is not None:
                test_id = line["test_id"]

                if not test_id in buffer:
                    buffer[test_id] = []

                for _t in skip_for_buffer[0]:
                    if test_id.startswith(_t + sep):
                        skip = True
                        break
                if not skip:
                    buffer[test_id].append(line)

            formatter = formatters.get(line["message_keyword"], None)
            if skip:
                line = None
            elif formatter:
                if formatter[0] is format_input and show_input is False:
                    line = None
                else:
                    line = formatter[0](
                        line,
                        *formatter[1:],
                        dump_transform,
                        buffer,
                        skip_for_dump,
                        skip_for_buffer,
                        only_new,
                    )
            else:
                line = None
        line = yield line



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import textwrap

from argparse import ArgumentParser as ArgumentParserBase

from .common import epilog
from .common import description
from .common import HelpFormatter
from .handlers.transform.handler import Handler as transform_handler
from .handlers.document.handler import Handler as document_handler
from .handlers.requirement.handler import Handler as requirement_handler
from .handlers.report.handler import Handler as report_handler
from .handlers.show.handler import Handler as show_handler
from .handlers.log import Handler as log_handler
from .handlers.run import Handler as run_handler

from .type import onoff as onoff_type

try:
    from .handlers.snapshot.handler import Handler as snapshot_handler
except Exception as e:
    snapshot_handler = None

try:
    from testflows.database.cli.handler import Handler as database_handler
except:
    database_handler = None

try:
    from testflows.enterprise._core.cli.handler import Handler as enterprise_handler
    from testflows.enterprise._core.cli.parser import (
        argument_parser as enterprise_parser,
    )
except:
    enterprise_handler = None
    enterprise_parser = None

from testflows._core import __version__, __license__


class ArgumentParser(ArgumentParserBase):
    """Customized argument parser."""

    def __init__(self, *args, **kwargs):
        description_prog = kwargs.pop("description_prog", None)
        kwargs["epilog"] = kwargs.pop("epilog", epilog())
        kwargs["description"] = description(
            textwrap.dedent(kwargs.pop("description", None) or ""),
            prog=description_prog,
        )
        kwargs["formatter_class"] = kwargs.pop("formatter_class", HelpFormatter)
        return super(ArgumentParser, self).__init__(*args, **kwargs)


parser = ArgumentParser(prog="tfs")

parser.add_argument(
    "--debug",
    dest="debug",
    action="store_true",
    help="enable debugging mode",
    default=False,
)
parser.add_argument(
    "--no-colors",
    dest="no_colors",
    action="store_true",
    help="disable terminal color highlighting",
    default=False,
)
parser.add_argument(
    "--show-skipped",
    dest="show_skipped",
    action="store_true",
    help="show skipped tests, default: False",
    default=False,
)
parser.add_argument(
    "--trim-results",
    dest="trim_results",
    type=onoff_type,
    help="enable or disable trimming of result messages, default: on",
    metavar=onoff_type.metavar,
    default="on",
)
parser.add_argument("-v", "--version", action="version", version=f"{__version__}")
parser.add_argument(
    "--license",
    action="version",
    help="show program's license and exit",
    version=f"{__license__}",
)
parser.add_argument(
    "--profile",
    action="store_true",
    help="enable profiler",
    default=False,
)

if enterprise_parser:
    enterprise_parser(parser)

commands = parser.add_subparsers(
    title="commands", metavar="command", description=None, help=None
)
commands.required = True

run_handler.add_command(commands)
log_handler.add_command(commands)
show_handler.add_command(commands)
report_handler.add_command(commands)
transform_handler.add_command(commands)
requirement_handler.add_command(commands)
if snapshot_handler:
    snapshot_handler.add_command(commands)
if database_handler:
    database_handler.add_command(commands)
document_handler.add_command(commands)

if enterprise_handler:
    enterprise_handler.add_command(commands)



# Copyright 2023 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows._core.cli.arg.handlers.snapshot.rewrite import (
    Handler as rewrite_handler,
)


class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser(
            "snapshots",
            help="snapshots processing",
            epilog=epilog(),
            description="Work with snapshots.",
            formatter_class=HelpFormatter,
        )

        snapshot_commands = parser.add_subparsers(
            title="commands", metavar="command", description=None, help=None
        )
        snapshot_commands.required = True
        rewrite_handler.add_command(snapshot_commands)



# Copyright 2023 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import itertools

from .covering_array import CoveringArray, CoveringArrayError

product = itertools.product
permutations = itertools.permutations
binomial = math.comb


def combinations(iterable, r, with_replacement=False):
    """Return successive r-length combinations of elements in the iterable.

    By default, returns combinations without replacement
    unless `with_replacement` is set to `True`.

    For example,

    without replacement

    ```
    combinations(range(4), 3) --> (0,1,2), (0,1,3), (0,2,3), (1,2,3)
    ```

    with replacement

    ```
    combinations(range(4), 3, with_replacement=True) -->
    [(0, 0, 0), (0, 0, 1), (0, 0, 2), (0, 0, 3), (0, 1, 1), (0, 1, 2),
     (0, 1, 3), (0, 2, 2), (0, 2, 3), (0, 3, 3), (1, 1, 1), (1, 1, 2),
     (1, 1, 3), (1, 2, 2), (1, 2, 3), (1, 3, 3), (2, 2, 2), (2, 2, 3),
     (2, 3, 3), (3, 3, 3)]
    ```

    :param iterable: iterable
    :param r: r-length
    :param with_replacement: with replacement, default: `False`
    """
    if with_replacement:
        return itertools.combinations_with_replacement(iterable, r)
    return itertools.combinations(iterable, r)



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# to the end flag
TE = 1 << 0
# utility test flag
UT = 1 << 1
# skip flag
SKIP = 1 << 2
# expected OK
EOK = 1 << 3
# expected Fail
EFAIL = 1 << 4
# expected Error
EERROR = 1 << 5
# expected Skip
ESKIP = 1 << 6
# crossed out ok
XOK = 1 << 7
# crossed out fail
XFAIL = 1 << 8
# crossed out error
XERROR = 1 << 9
# crossed out null
XNULL = 1 << 10
# not counted fail
FAIL_NOT_COUNTED = 1 << 11
# not counted error
ERROR_NOT_COUNTED = 1 << 12
# not counted null
NULL_NOT_COUNTED = 1 << 13
# pause before test execution
PAUSE_BEFORE = 1 << 14
# pause before short form
PAUSE = PAUSE_BEFORE
# pause after test execution
PAUSE_AFTER = 1 << 15
# manual test
MANUAL = 1 << 16
# auto test
AUTO = 1 << 17
# report flag
REPORT = 1 << 26
# documentation
DOCUMENT = 1 << 27
# mandatory test
MANDATORY = 1 << 28
# parallel test
PARALLEL = 1 << 29
# disable parallel tests
NO_PARALLEL = 1 << 30
# clear flags mask
CLEAR = 1 << 31
# async test
ASYNC = 1 << 32
# repeated test flag
REPEATED = 1 << 33
# not repeatable test
NOT_REPEATABLE = 1 << 34
# retried test flag
RETRIED = 1 << 35
# last retry flag
LAST_RETRY = 1 << 36
# pause on passing result
PAUSE_ON_PASS = 1 << 37
# pause on failing result
PAUSE_ON_FAIL = 1 << 38
# remote test
REMOTE = 1 << 39
# setup
SETUP = 1 << 40
# cleanup
CLEANUP = 1 << 41
# retry
RETRY = 1 << 42
# nested retry
NESTED_RETRY = 1 << 43
# expected result
ERESULT = EOK | EFAIL | ESKIP | EERROR
# expected any result
EANY = EOK | EFAIL | ESKIP
# crossed out result
XRESULT = XOK | XFAIL | XERROR | XNULL
# not counted
NOT_COUNTED = FAIL_NOT_COUNTED
NOT_COUNTED_ANY = FAIL_NOT_COUNTED | ERROR_NOT_COUNTED | NULL_NOT_COUNTED
# cumulative flags
CFLAGS = (
    UT
    | MANDATORY
    | MANUAL
    | PARALLEL
    | NO_PARALLEL
    | NOT_REPEATABLE
    | RETRY
    | NESTED_RETRY
    | SETUP
    | CLEANUP
)


class Flags(object):
    """Test flags."""

    all = [
        TE,
        UT,
        SKIP,
        EOK,
        EFAIL,
        EERROR,
        ESKIP,
        XOK,
        XFAIL,
        XERROR,
        XNULL,
        FAIL_NOT_COUNTED,
        ERROR_NOT_COUNTED,
        NULL_NOT_COUNTED,
        PAUSE_BEFORE,
        PAUSE_AFTER,
        MANUAL,
        AUTO,
        REPORT,
        DOCUMENT,
        MANDATORY,
        PARALLEL,
        NO_PARALLEL,
        CLEAR,
        ASYNC,
        REPEATED,
        NOT_REPEATABLE,
        RETRIED,
        LAST_RETRY,
        PAUSE_ON_PASS,
        PAUSE_ON_FAIL,
        REMOTE,
        SETUP,
        CLEANUP,
        RETRY,
        NESTED_RETRY,
    ]
    all_str = [
        "TE",
        "UT",
        "SKIP",
        "EOK",
        "EFAIL",
        "EERROR",
        "ESKIP",
        "XOK",
        "XFAIL",
        "XERROR",
        "XNULL",
        "FAIL_NOT_COUNTED",
        "ERROR_NOT_COUNTED",
        "NULL_NOT_COUNTED",
        "PAUSE_BEFORE",
        "PAUSE_AFTER",
        "MANUAL",
        "AUTO",
        "REPORT",
        "DOCUMENT",
        "MANDATORY",
        "PARALLEL",
        "NO_PARALLEL",
        "CLEAR",
        "ASYNC",
        "REPEATED",
        "NOT_REPEATABLE",
        "RETRIED",
        "LAST_RETRY",
        "PAUSE_ON_PASS",
        "PAUSE_ON_FAIL",
        "REMOTE",
        "SETUP",
        "CLEANUP",
        "RETRY",
        "NESTED_RETRY",
    ]

    def __init__(self, flags=0):
        if flags is None:
            flags = 0
        if type(flags) is str:
            self.flags = 0
            if flags:
                for flag in [self.all[self.all_str.index(f)] for f in flags.split("|")]:
                    self.flags |= flag
        else:
            self.flags = int(flags)

    def __str__(self):
        l = [self.all_str[self.all.index(f)] for f in self.all if self.flags & f == f]
        return "|".join(l) or ""

    def __repr__(self):
        return "Flags('" + str(self) + "')"

    def __bool__(self):
        return bool(self.flags)

    def __int__(self):
        return self.flags

    def __and__(self, o):
        return Flags(self.flags & int(Flags(o)))

    def __rand__(self, o):
        return self.__and__(o)

    def __or__(self, o):
        return Flags(self.flags | int(Flags(o)))

    def __ror__(self, o):
        return self.__or__(o)

    def __xor__(self, o):
        return Flags(self.flags ^ int(Flags(o)))

    def __rxor__(self, o):
        return self.__xor__(o)

    def __invert__(self):
        return Flags(~self.flags)

    def __contains__(self, o):
        return bool(self & o)

    def __eq__(self, o):
        return self.flags == int(Flags(o))

    def __ne__(self, o):
        return not self == o

    def keys(self):
        return ["flags"]

    def __getitem__(self, key):
        if key == "flags":
            return self
        raise KeyError(key)

    def __call__(self, func):
        setattr(func, "flags", self)
        return func



# Copyright 2023 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import time
import signal
import threading
import subprocess

import testflows._core.cli.arg.type as argtype

from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase

description = """Run test program.

If either '--stdout' or '--stderr' is specified then '--no-colors'
is automatically applied to the test program.
You can overwrite this behavior by passing `--no-colors off` option
to the test program.

Examples:

Write pid, stdout and stderr to files.
    tfs run --pid run.pid --stdout run.out --stderr run.err test.py -- -o classic

Write pid, stdout and stderr to files but silence output to the terminal.
    tfs run --no-output --pid run.pid --stdout run.out --stderr run.err test.py -- -o classic

Redirect stdout to file but turn colors back on.
    tfs run --stdout run.out --stderr run.err test.py -- -o classic --no-colors off

Write exitcode to file
    tfs run --exitcode run.exitcode test.py
"""


class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser(
            "run",
            help="run test program",
            epilog=epilog(),
            description=description,
            formatter_class=HelpFormatter,
        )

        parser.add_argument("--pid", metavar="file", type=str, help="pid file")
        parser.add_argument(
            "--exitcode", metavar="file", type=str, help="exit code file"
        )

        parser.add_argument(
            "--stdout", metavar="file", type=str, help="file where to redirect stdout"
        )

        parser.add_argument(
            "--stderr", metavar="file", type=str, help="file where to redirect stderr"
        )

        parser.add_argument(
            "-q",
            "--no-output",
            action="store_true",
            default=False,
            help="disable output to the terminal from redirected stdout",
        )

        parser.add_argument(
            "program", metavar="program", type=argtype.path, help="test program to run"
        )

        parser.add_argument(
            "args", metavar="args", nargs="*", type=str, help="test program arguments"
        )

        parser.set_defaults(func=cls())

    def _reader(self, filename, out, process, flush=True, timeout=0.1):
        """Read contents of file and write to the specified output."""
        with open(filename, "r") as fd:
            while True:
                try:
                    out.write(fd.read())
                    if flush:
                        out.flush()
                finally:
                    if process.returncode is not None:
                        break
                    time.sleep(timeout)

    def handle(self, args):
        command = ["python3", os.path.abspath(args.program)]

        stdout = sys.stdout
        if args.stdout:
            stdout = open(args.stdout, "w")

        stderr = sys.stderr
        if args.stderr:
            stderr = open(args.stderr, "w")

        if args.stdout or args.stderr:
            command.append("--no-colors")

        command += args.args

        process = subprocess.Popen(
            command, stdin=None, stderr=stderr, stdout=stdout, close_fds=True
        )

        if args.pid:
            with open(args.pid, "w") as fd:
                fd.write(f"{os.getpid()}\n")

        stop_event = threading.Event()
        stdout_reader = None
        stderr_reader = None

        try:
            if not args.no_output:
                if args.stdout:
                    stdout_reader = threading.Thread(
                        target=self._reader,
                        args=(args.stdout, sys.stdout, process, True),
                    )
                    stdout_reader.start()
            if args.stderr:
                stderr_reader = threading.Thread(
                    target=self._reader, args=(args.stderr, sys.stderr, process)
                )
                stderr_reader.start()

            while True:
                try:
                    if process.poll() is not None:
                        break
                    if (
                        subprocess.call(
                            ["kill", "-0", f"{process.pid}"],
                            stderr=subprocess.DEVNULL,
                            stdout=subprocess.DEVNULL,
                        )
                        != 0
                    ):
                        break
                    time.sleep(0.25)
                except KeyboardInterrupt:
                    process.send_signal(signal.SIGINT)
                except BaseException:
                    process.terminate()
        finally:
            try:
                if stdout_reader is not None:
                    stdout_reader.join()
                if stderr_reader is not None:
                    stderr_reader.join()
            finally:
                sys.stdout.flush()
                sys.stderr.flush()
                if args.exitcode:
                    with open(args.exitcode, "w") as fd:
                        fd.write(f"{process.returncode}\n")
                sys.exit(process.returncode if process.returncode is not None else -1)



# Copyright 2020 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from testflows._core.name import match, filter, matchcase, translate, normname, join
from testflows._core.name import (
    isabs,
    absname,
    basename,
    parentname,
    depth,
    split,
    relname,
    sep,
    escape,
    clean,
    unclean,
)
from testflows._core.name import commonprefix, commonname, normcase



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import io
import sys
import csv
import argparse

from argparse import ArgumentTypeError
from collections import namedtuple
from testflows._core.exceptions import exception
from testflows._core.compress import CompressedFile
from testflows._core.objects import Repeat, Retry
from testflows._core.tracing import logging

import testflows._core.contrib.rsa as rsa

KeyValue = namedtuple("KeyValue", "key value")
NoneValue = "__none__"


class FileType(object):
    def __init__(self, mode="r", bufsize=-1, encoding=None, errors=None, safe=False):
        self._mode = mode
        self._bufsize = bufsize
        self._encoding = encoding
        self._errors = errors
        self._safe = safe

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == "-":
            if "r" in self._mode:
                if "b" in self._mode:
                    return sys.stdin.buffer
                return sys.stdin
            elif "w" in self._mode:
                if "b" in self._mode:
                    return sys.stdout.buffer
                return sys.stdout
            else:
                msg = argparse._('argument "-" with mode %r') % self._mode
                raise ValueError(msg)

        # all other arguments are used as file names
        try:
            if self._safe and os.path.exists(string):
                raise FileExistsError(
                    f"File '{string}' already exists, please delete it first."
                )

            return open(string, self._mode, self._bufsize, self._encoding, self._errors)
        except OSError as e:
            message = argparse._("can't open '%s': %s")
            raise ArgumentTypeError(message % (string, e))

    def __repr__(self):
        args = self._mode, self._bufsize
        kwargs = [("encoding", self._encoding), ("errors", self._errors)]
        args_str = ", ".join(
            [repr(arg) for arg in args if arg != -1]
            + ["%s=%r" % (kw, arg) for kw, arg in kwargs if arg is not None]
        )
        return "%s(%s)" % (type(self).__name__, args_str)


class LogFileType(object):
    def __init__(self, mode="r", bufsize=-1, encoding=None, errors=None):
        self._mode = mode
        self._encoding = encoding
        self._errors = errors

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == "-":
            if "r" in self._mode:
                fp = CompressedFile(sys.stdin.buffer, self._mode)
                if self._encoding:
                    return io.TextIOWrapper(fp, self._encoding, self._errors)
                return fp
            elif "w" in self._mode:
                fp = CompressedFile(sys.stdout.buffer, self._mode)
                if self._encoding:
                    return io.TextIOWrapper(fp, self._encoding, self._errors)
                return fp
            else:
                msg = argparse._('argument "-" with mode %r') % self._mode
                raise ValueError(msg)

        # all other arguments are used as file names
        try:
            fp = CompressedFile(string, self._mode)
            if self._encoding:
                return io.TextIOWrapper(fp, self._encoding, self._errors)
            return fp
        except OSError as e:
            message = argparse._("can't open '%s': %s")
            raise ArgumentTypeError(message % (string, e))

    def __repr__(self):
        args = self._mode
        kwargs = [("encoding", self._encoding), ("errors", self._errors)]
        args_str = ", ".join(
            [repr(arg) for arg in args if arg != -1]
            + ["%s=%r" % (kw, arg) for kw, arg in kwargs if arg is not None]
        )
        return "%s(%s)" % (type(self).__name__, args_str)


def path(p, special=""):
    if p in special or []:
        return p
    if not os.path.exists(os.path.abspath(p)):
        raise ArgumentTypeError(f"path does not exist: '{os.path.abspath(p)}'")
    return p


def file(*args, **kwargs):
    """File type."""
    return FileType(*args, **kwargs)


def logfile(*args, **kwargs):
    """Log file type."""
    return LogFileType(*args, **kwargs)


def rsa_private_key_pem_file(p):
    """RSA private key PEM file type."""
    with open(p, mode="rb") as pem_file:
        return rsa.PrivateKey.load_pkcs1(pem_file.read())


def key_value(s, sep="="):
    """Parse a key, value pair using a seperator (default: '=')."""
    if sep not in s:
        raise ArgumentTypeError(f"invalid format of key{sep}value")
    key, value = s.split(sep, 1)
    return KeyValue(key.strip(), value.strip())


def count(value):
    try:
        value = int(value)
        assert value >= 0
    except:
        raise ArgumentTypeError(f"{value} is not a positive number")
    return value


def repeat(value):
    try:
        fields = list(csv.reader([value], "unix"))[-1]
        until = "fail"
        if len(fields) > 2:
            pattern, count, until = fields
        else:
            pattern, count = fields
        option = Repeat(count=count, pattern=pattern, until=until)
    except Exception as e:
        raise ArgumentTypeError(f"'{value}' is invalid")
    return option


def retry(value):
    try:
        fields = list(csv.reader([value], "unix"))[-1]
        jitter = None  # can't be specified
        if len(fields) < 2:
            raise ValueError("needs at least 2 values")
        pattern, count, timeout, delay, backoff = [
            *fields,
            *["", None, None, 0, 1][len(fields) :],
        ]
        if not pattern:
            pattern = ""
        if not count:
            count = None
        if not timeout:
            timeout = None
        if not delay:
            delay = 0
        if not backoff:
            backoff = 1
        option = Retry(
            count=count,
            timeout=timeout,
            delay=delay,
            backoff=backoff,
            jitter=jitter,
            pattern=pattern,
        )
    except Exception as e:
        raise ArgumentTypeError(f"'{value}' is invalid")
    return option


def tags_filter(value):
    try:
        type, cvstags = value.split(":", 1)
        assert type in ["test", "suite", "module", "feature", "scenario", "any"]
        tags = list(csv.reader([cvstags], "unix"))[-1]
        if type == "scenario":
            type = "test"
        elif type == "feature":
            type = "suite"
        option = (type, tuple(tags))
    except Exception as e:
        raise ArgumentTypeError(f"'{value}' is invalid")
    return option


def onoff(value):
    if value.lower() in ["yes", "1", "on"]:
        return True
    elif value.lower() in ["no", "0", "off"]:
        return False
    elif value == NoneValue:
        return NoneValue
    raise ArgumentTypeError(f"'{value}' is invalid")


onoff.metavar = (
    str(set(["yes", "1", "on", "no", "0", "off"])).replace(" ", "").replace("'", "")
)


def trace_level(value):
    if value.lower() in ["debug", "info", "warning", "error", "critical"]:
        return getattr(logging, value.upper())
    else:
        raise ArgumentTypeError(f"'{value}' is invalid")


trace_level.choices = ["debug", "info", "warning", "error", "critical"]
trace_level.metavar = str(set(trace_level.choices)).replace(" ", "").replace("'", "")



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows._core.cli.arg.handlers.transform.nice import Handler as nice_handler
from testflows._core.cli.arg.handlers.transform.pnice import Handler as pnice_handler
from testflows._core.cli.arg.handlers.transform.brisk import Handler as brisk_handler
from testflows._core.cli.arg.handlers.transform.plain import Handler as plain_handler
from testflows._core.cli.arg.handlers.transform.short import Handler as short_handler
from testflows._core.cli.arg.handlers.transform.dots import Handler as dots_handler
from testflows._core.cli.arg.handlers.transform.compact import (
    Handler as compact_handler,
)
from testflows._core.cli.arg.handlers.transform.slick import Handler as slick_handler
from testflows._core.cli.arg.handlers.transform.classic import (
    Handler as classic_handler,
)
from testflows._core.cli.arg.handlers.transform.manual import Handler as manual_handler
from testflows._core.cli.arg.handlers.transform.fails import Handler as fails_handler
from testflows._core.cli.arg.handlers.transform.new_fails import (
    Handler as new_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.brisk_fails import (
    Handler as brisk_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.brisk_new_fails import (
    Handler as brisk_new_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.plain_fails import (
    Handler as plain_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.plain_new_fails import (
    Handler as plain_new_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.nice_fails import (
    Handler as nice_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.nice_new_fails import (
    Handler as nice_new_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.pnice_fails import (
    Handler as pnice_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.pnice_new_fails import (
    Handler as pnice_new_fails_handler,
)
from testflows._core.cli.arg.handlers.transform.raw import Handler as raw_handler
from testflows._core.cli.arg.handlers.transform.compress import (
    Handler as compress_handler,
)
from testflows._core.cli.arg.handlers.transform.decompress import (
    Handler as decompress_handler,
)

try:
    from testflows.enterprise._core.cli.transform.handler import (
        Handler as enterprise_handler,
    )
except:
    enterprise_handler = None


class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser(
            "transform",
            help="log file transformation",
            epilog=epilog(),
            description="Transform log files.",
            formatter_class=HelpFormatter,
        )

        transform_commands = parser.add_subparsers(
            title="commands", metavar="command", description=None, help=None
        )
        transform_commands.required = True
        raw_handler.add_command(transform_commands)
        nice_handler.add_command(transform_commands)
        pnice_handler.add_command(transform_commands)
        brisk_handler.add_command(transform_commands)
        plain_handler.add_command(transform_commands)
        short_handler.add_command(transform_commands)
        slick_handler.add_command(transform_commands)
        classic_handler.add_command(transform_commands)
        manual_handler.add_command(transform_commands)
        fails_handler.add_command(transform_commands)
        new_fails_handler.add_command(transform_commands)
        brisk_fails_handler.add_command(transform_commands)
        brisk_new_fails_handler.add_command(transform_commands)
        plain_fails_handler.add_command(transform_commands)
        plain_new_fails_handler.add_command(transform_commands)
        nice_fails_handler.add_command(transform_commands)
        nice_new_fails_handler.add_command(transform_commands)
        pnice_fails_handler.add_command(transform_commands)
        pnice_new_fails_handler.add_command(transform_commands)
        dots_handler.add_command(transform_commands)
        compact_handler.add_command(transform_commands)
        compress_handler.add_command(transform_commands)
        decompress_handler.add_command(transform_commands)
        if enterprise_handler is not None:
            enterprise_handler.add_command(transform_commands)



# Copyright 2023 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from testflows._core.cli.text import secondary
from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows.snapshots.v1 import rewrite_snapshot


class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser(
            "rewrite",
            help="rewrite snapshot",
            epilog=epilog(),
            description=(
                "Rewrite snapshot files.\n\n"
                "For example:\n"
                "  tfs snapshots rewrite *.snapshot"
            ),
            formatter_class=HelpFormatter,
        )

        parser.add_argument(
            "input",
            metavar="input",
            type=str,
            nargs="+",
            help="input file or files",
        )

        parser.set_defaults(func=cls())

    def handle(self, args):
        for filename in args.input:
            try:
                rewrite_snapshot(filename)
            except ValueError as exc:
                print(secondary(f"skipping: {exc}", eol=""), file=sys.stderr)
            else:
                print(secondary(f"rewrote: {filename}", eol=""))



# Copyright 2021 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# to the end flag
from concurrent.futures import Future as _FutureBase


class Future(_FutureBase):
    """concurrent.futures.Future that
    fixes __get_result() when exception
    implements its own custom __bool__()
    """

    def __get_result(self):
        if self._exception is not None:
            try:
                raise self._exception
            finally:
                # Break a reference cycle with the exception in self._exception
                self = None
        else:
            return self._result



"""Code is based on Python3.6 posixpath.py.
"""

import re
import functools

sep = "/"
empty = ""
dot = "."
dotdot = ".."
pardir = dotdot
curdir = dot


def clean(name):
    """Clean name by converting all special pattern and regex
    characters to their UTF-8 replacements to avoid conflicts."""
    return name.translate(
        str.maketrans(
            {
                # '/' is not allowed just like in Unix file names
                # so convert any '/' to U+2215 division slash
                sep: "\u2215",  # path
                '"': "\uFF02",  # bash
                "'": "\uFF07",  # bash
                "$": "\uFE69",  # bash
                "\\": "\uFE68",  # bash special symbol
                "[": "\uFF3B",  # pattern / regex
                "]": "\uFF3D",  # pattern / regex
                "*": "\uFF0A",  # pattern / regex
                "?": "\uFE16",  # pattern / regex
                ":": "\uFE55",  # pattern
                "!": "\uFE15",  # bash
                ".": "\u2024",  # regex
                "^": "\u02C4",  # regex
                "+": "\uFF0B",  # regex
                "{": "\uFF5B",  # regex
                "}": "\uFF5D",  # regex
                "|": "\u2160",  # regex
                "(": "\uFF08",  # regex
                ")": "\uFF09",  # regex
            }
        )
    )


def unclean(name):
    """Try to revert name cleaning by converting all converted special pattern and regex
    characters from their UTF-8 replacements back to the original characters.

    Warning: this function will not work if the name had a mix of restricted characters
    and UTF-8 replacement characters as all UTF-8 replacement characters will be converted.
    """
    return name.translate(
        str.maketrans(
            {
                "\u2215": sep,  # path
                "\uFF02": '"',  # bash
                "\uFF07": "'",  # bash
                "\uFE69": "$",  # bash
                "\uFE68": "\\",  # bash special symbol
                "\uFF3B": "[",  # pattern / regex
                "\uFF3D": "]",  # pattern / regex
                "\uFF0A": "*",  # pattern / regex
                "\uFE16": "?",  # pattern / regex
                "\uFE55": ":",  # pattern
                "\uFE15": "!",  # bash
                "\u2024": ".",  # regex
                "\u02C4": "^",  # regex
                "\uFF0B": "+",  # regex
                "\uFF5B": "{",  # regex
                "\uFF5D": "}",  # regex
                "\u2160": "|",  # regex
                "\uFF08": "(",  # regex
                "\uFF09": ")",  # regex
            }
        )
    )


def escape(name):
    """Escape special pattern characters in name."""
    # the closing square bracket ']' does not need to be escaped
    # as pattern is parsed from left to right
    return (
        name.replace("[", rf"[\[]")
        .replace("*", "[*]")
        .replace(":", "[:]")
        .replace("?", "[?]")
    )


def match(name, pat, prefix=False):
    """Test whether FILENAME matches PATTERN.

    Patterns are modified Unix shell style:

    *       matches everything
    :       matches everything but path separator '/'
    ?       matches any single character
    [seq]   matches any character in seq
    [!seq]  matches any char not in seq

    An initial period in FILENAME is not special.
    Both FILENAME and PATTERN are first case-normalized
    if the operating system requires it.
    If you don't want this, use matchcase(FILENAME, PATTERN).
    """
    name = normcase(name)
    pat = normcase(pat)
    return matchcase(name, pat, prefix=prefix)


def filter(names, pat, prefix=False):
    """Return the subset of the list NAMES that match PAT."""
    result = []
    match = _compile_pattern(pat, prefix=prefix).match
    # normcase is NOP. Optimize it away from the loop.
    for name in names:
        if match(name):
            result.append(name)
    return result


def matchcase(name, pat, prefix=False):
    """Test whether FILENAME matches PATTERN, including case.

    Patterns are modified Unix shell style:

    *       matches everything
    :       matches everything but path separator '/'
    ?       matches any single character
    [seq]   matches any character in seq
    [!seq]  matches any char not in seq

    This is a version of fnmatch() which doesn't case-normalize
    its arguments.
    """
    match = _compile_pattern(pat, prefix=prefix).match
    return match(name) is not None


@functools.lru_cache(maxsize=256, typed=True)
def _compile_pattern(pat, prefix=False):
    res = translate(pat, prefix=prefix)
    return re.compile(res)


def translate(pat, prefix=False):
    """Translate a shell PATTERN to a regular expression.

    If prefix is set to True, then translates a shell
    PATTERN to a regular expression that matches
    longest prefix of the path.

    For example the pattern  'A/B/C' would match
    'A', 'A/B' , 'A/B/C'
    as all are prefixes of the full match 'A/B/C'.

    However it would not match any path prefix
    that ends with '/' such as 'A/', or 'A/B/'
    as these paths are treated as incomplete.
    If you need to match such path you need to
    rstrip() the '/' before matching.

    There is no way to quote meta-characters.
    """
    if prefix:
        return _translate_prefix(pat)
    return _translate(pat)


def _translate(pat):
    """Translate a shell PATTERN to a regular expression.

    There is no way to quote meta-characters.
    """
    i, n = 0, len(pat)
    res = ""
    while i < n:
        c = pat[i]
        i = i + 1
        if c == "*":
            res = res + ".*"
        elif c == ":":
            res = res + "[^/]+"
        elif c == "?":
            res = res + "."
        elif c == "[":
            j = i
            if j < n and pat[j] == "!":
                j = j + 1
            if j < n and pat[j] == "]":
                j = j + 1
            while j < n and pat[j] != "]":
                j = j + 1
            if j >= n:
                res = res + "\\["
            else:
                stuff = pat[i:j].replace("\\", "\\\\")
                i = j + 1
                if stuff[0] == "!":
                    stuff = "^" + stuff[1:]
                elif stuff[0] == "^":
                    stuff = "\\" + stuff
                res = "%s[%s]" % (res, stuff)
        else:
            res = res + re.escape(c)
    return r"(?s:%s)\Z" % res


def _translate_prefix(pat):
    """Translate a shell PATTERN to a regular expression
    that matches longest prefix of the path.

    For example the pattern  'A/B/C' would match
    'A', 'A/B' , 'A/B/C'
    as all are prefixes of the full match 'A/B/C'.

    However it would not match any path prefix
    that ends with '/' such as 'A/', or 'A/B/'
    as these paths are treated as incompleted.
    If you need to match such path you need to
    rstrip() the '/' before matching.

    There is no way to quote meta-characters.
    """

    i, n = 0, len(pat)
    res = ""
    level = 0
    while i < n:
        c = pat[i]
        i = i + 1
        if c == "/":
            level += 1
            res = res + "(/"
        elif c == "*":
            level += 1
            res = res + "(.*"
        elif c == ":":
            level += 1
            res = res + "([^/]+"
        elif c == "?":
            level += 1
            res = res + "(."
        elif c == "[":
            has = False
            j = i
            if j < n and pat[j] == "!":
                j = j + 1
            if j < n and pat[j] == "]":
                j = j + 1
            while j < n and pat[j] != "]":
                if pat[j] == "/":
                    has = True
                j = j + 1
            if j >= n:
                res = res + "\\["
            else:
                stuff = pat[i:j].replace("\\", "\\\\")
                i = j + 1
                if stuff[0] == "!":
                    stuff = "^" + stuff[1:]
                elif stuff[0] == "^":
                    stuff = "\\" + stuff
                if has:
                    level += 1
                    res = "%s([%s]" % (res, stuff)
                else:
                    res = "%s[%s]" % (res, stuff)
        else:
            res = res + re.escape(c)
    for l in range(level):
        res = res + ")?"
    return r"(?s:%s)\Z" % res


# Normalize a path, e.g. A//B, A/./B and A/foo/../B all become A/B.
def normname(name):
    comps = name.split(sep)
    new_comps = []
    initial_slashes = name.startswith(sep)
    for comp in comps:
        if comp in (empty, dot):
            continue
        if (
            comp != dotdot
            or (not initial_slashes and not new_comps)
            or (new_comps and new_comps[-1] == dotdot)
        ):
            new_comps.append(comp)
        elif new_comps:
            new_comps.pop()
    comps = new_comps
    name = sep.join(comps)
    if initial_slashes:
        name = sep * initial_slashes + name
    return name or dot


# Join pathnames.
# Ignore the previous parts if a part is absolute.
# Insert a '/' unless the first part is empty or already ends in '/'.
def join(a, *p):
    """Join two or more name components, inserting '/' as needed.
    If any component is an absolute name, all previous name components
    will be discarded.  An empty last part will result in a name that
    ends with a separator."""
    name = a
    if not p:
        name[:0] + sep
    for b in p:
        if b.startswith(sep):
            name = b
        elif not name or name.endswith(sep):
            name += b
        else:
            name += sep + b
    return name


def isabs(name):
    """Test whether a name is absolute."""
    return name.startswith(sep)


def absname(n, at):
    """Return an absolute name relative
    to the name specified in the `at`.
    """
    if not isabs(n):
        n = join(at, n)
    return normname(n)


# Return the tail (basename) part of a name, same as split(name)[1].
def basename(name):
    """Returns the final component of a name."""
    i = name.rfind(sep) + 1
    return name[i:]


# Return the head (dirname) part of a name, same as split(name)[0].
def parentname(name):
    """Returns the directory component of a name."""
    i = name.rfind(sep) + 1
    head = name[:i]
    if head and head != sep * len(head):
        head = head.rstrip(sep)
    return head


def depth(name):
    """Return depth of the path."""
    return name.count(sep)


# Split a path in head (everything up to the last '/') and tail (the
# rest).  If the path ends in '/', tail will be empty.  If there is no
# '/' in the path, head  will be empty.
# Trailing '/'es are stripped from head unless it is the root.
def split(name):
    """Split a name.  Returns tuple "(head, tail)" where "tail" is
    everything after the final slash.  Either part may be empty.
    """
    i = name.rfind(sep) + 1
    head, tail = name[:i], name[i:]
    if head and head != sep * len(head):
        head = head.rstrip(sep)
    return head, tail


def relname(name, at, start=None):
    """Return a relative version of a name
    relative to the `at`.
    """
    if not name:
        raise ValueError("no name specified")

    if start is None:
        start = curdir

    start_list = [x for x in absname(start, at).split(sep) if x]
    name_list = [x for x in absname(name, at).split(sep) if x]
    # Work out how much of the filepath is shared by start and path.
    i = len(commonprefix([start_list, name_list]))

    rel_list = [pardir] * (len(start_list) - i) + name_list[i:]
    if not rel_list:
        return curdir
    return join(*rel_list)


# Return the longest prefix of all list elements.
def commonprefix(m):
    """Given a list of pathnames, returns the longest common leading component."""
    if not m:
        return ""
    # Some people pass in a list of pathname parts to operate in an OS-agnostic
    # fashion; don't try to translate in that case as that's an abuse of the
    # API and they are already doing what they need to be OS-agnostic and so
    # they most likely won't be using an os.PathLike object in the sublists.
    if not isinstance(m[0], (list, tuple)):
        m = tuple(m)
    s1 = min(m)
    s2 = max(m)
    for i, c in enumerate(s1):
        if c != s2[i]:
            return s1[:i]
    return s1


# Return the longest common sub-path of the sequence of paths given as input.
# The paths are not normalized before comparing them (this is the
# responsibility of the caller). Any trailing separator is stripped from the
# returned path.
def commonname(names):
    """Given a sequence of names, returns the longest common sub-name."""
    if not names:
        raise ValueError("names is an empty sequence")

    split_names = [name.split(sep) for name in names]

    try:
        (isabs,) = set(n[:1] == sep for n in names)
    except ValueError:
        raise ValueError("can't mix absolute and relative names") from None

    split_names = [[c for c in s if c and c != curdir] for s in split_names]
    s1 = min(split_names)
    s2 = max(split_names)
    common = s1
    for i, c in enumerate(s1):
        if c != s2[i]:
            common = s1[:i]
            break

    prefix = sep if isabs else sep[:0]
    return prefix + sep.join(common)


def normcase(s):
    """Normalize case of pathname."""
    return s



# Copyright 2019 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tempfile
import testflows._core.cli.arg.type as argtype

from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
from testflows._core.document.toc import generate


class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser(
            "toc",
            help="generate table of contents",
            epilog=epilog(),
            description="Genarate table of contents for a document.",
            formatter_class=HelpFormatter,
        )

        parser.add_argument(
            "input",
            metavar="input",
            type=argtype.file("r", bufsize=1, encoding="utf-8"),
            nargs="?",
            help="input file, default: stdin",
            default="-",
        )
        parser.add_argument(
            "output",
            metavar="output",
            type=str,
            nargs="?",
            help="output file, default: stdout",
            default="-",
        )

        parser.add_argument(
            "--heading",
            metavar="name",
            type=str,
            default="Table of Contents",
            help="table of contents heading name, default: 'Table of Contents'",
        )
        parser.add_argument(
            "-u",
            "--update",
            action="store_true",
            default=False,
            help="output original document with an updated table of contents",
        )

        parser.set_defaults(func=cls())

    def handle(self, args):
        with tempfile.TemporaryFile("w+", encoding="utf-8") as temporary_file:
            generate(args.input, temporary_file, args.heading, args.update)
            output = argtype.file("w", encoding="utf-8")(args.output)
            temporary_file.flush()
            temporary_file.seek(0)
            output.write(temporary_file.read())
            output.flush()



# Copyright 2020 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import datetime
import testflows._core.cli.arg.type as argtype

from testflows._core.cli.arg.common import epilog
from testflows._core.cli.arg.common import HelpFormatter
from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase


class Handler(HandlerBase):
    @classmethod
    def add_command(cls, commands):
        parser = commands.add_parser(
            "requirements",
            help="software requirements specification",
            epilog=epilog(),
            description="Create new software requirements specification document.",
            formatter_class=HelpFormatter,
        )

        parser.add_argument(
            "output",
            metavar="output",
            type=argtype.file("w", bufsize=1, encoding="utf-8"),
            nargs="?",
            help="output file, default: stdout",
            default="-",
        )
        parser.add_argument(
            "--number",
            metavar="number",
            type=str,
            help="document number",
            default="001",
        )
        parser.add_argument(
            "--title",
            metavar="name",
            type=str,
            help="document title",
            default="Template",
        )
        parser.add_argument(
            "--author",
            metavar="name",
            type=str,
            help="name of the author",
            default="*[author]*",
        )
        parser.add_argument(
            "--date",
            metavar="date",
            type=str,
            help="document date",
            default=f"{datetime.datetime.now():%B %d, %Y}",
        )
        parser.set_defaults(func=cls())

    def handle(self, args):
        template_dir = os.path.join(
            os.path.dirname(__file__), "..", "..", "..", "..", "..", "document", "new"
        )

        with open(os.path.join(template_dir, "requirements.md")) as fd:
            template = fd.read()

        args.output.write(
            template.format(
                number=args.number, title=args.title, author=args.author, date=args.date
            )
        )



# Copyright (C) 2021-2022 by Ivan.
# This file is part of Snowflake package.
# Snowflake is released under the MIT License (see LICENSE).


from .snowflake import *


def version():
    return "0.0.3"



# Copyright 2021 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import inspect
import asyncio
import concurrent

from asyncio import *


main_event_loop = None


if main_event_loop is None:
    loop = asyncio.get_event_loop()
    main_event_loop = loop


async def async_await(coroutine):
    r = coroutine
    while inspect.isawaitable(r):
        r = await r
    return r


async def async_next(async_iterator):
    return await async_iterator.__anext__()


def is_running_in_event_loop():
    """Check if running in async event loop."""
    try:
        loop = asyncio.get_running_loop()
        if loop is main_event_loop:
            # don't consider a main event loop (if present)
            # as an event loop where we expect all tests to be async
            return False
        return True
    except RuntimeError:
        pass
    return False


class LoopMixin:
    def __init__(self, *args, loop=None, **kwargs):
        self._loop_mixin = loop
        if sys.version_info < (3, 10, 0):
            kwargs["loop"] = self._loop_mixin
        super(LoopMixin, self).__init__(*args, **kwargs)

    def _get_loop(self):
        if self._loop_mixin is not None:
            return self._loop_mixin
        if sys.version_info >= (3, 10, 0):
            return super(LoopMixin, self)._get_loop()
        return self._loop


class Future(LoopMixin, asyncio.Future):
    pass


class OptionalFuture(Future):
    """Future that will not complain about any
    exceptions not being retrieved.
    """

    def __del__(self):
        return


def wrap_future(future, *, loop=None, new_future=None):
    """Wrap concurrent.futures.Future object."""
    if isfuture(future):
        return future
    assert isinstance(
        future, concurrent.futures.Future
    ), f"concurrent.futures.Future is expected, got {future!r}"
    if loop is None:
        loop = asyncio.events.get_event_loop()
    if new_future is None:
        new_future = loop.create_future()
    asyncio.futures._chain_future(future, new_future)
    return new_future


class Event(LoopMixin, asyncio.Event):
    pass


class Queue(LoopMixin, asyncio.Queue):
    pass


class Lock(LoopMixin, asyncio.Lock):
    pass



# Copyright 2019-2022 Katteli Inc.
# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import traceback


def exception(exc_type=None, exc_value=None, exc_traceback=None):
    """Get exception string."""
    if (exc_type, exc_value, exc_traceback) == (None, None, None):
        exc_type, exc_value, exc_traceback = sys.exc_info()
    return "".join(
        traceback.format_exception(exc_type, exc_value, exc_traceback)
    ).rstrip()


class TestFlowsException(Exception):
    """Base exception class."""

    pass


class ResultException(TestFlowsException):
    """Result exception."""

    pass


class DummyTestException(TestFlowsException):
    """Dummy test exception."""

    pass


class TestIteration(TestFlowsException):
    """Repeat test."""

    def __init__(self, repeat, retry, *args, **kwargs):
        self.repeat = repeat
        self.retry = retry
        super(TestIteration, self).__init__(*args, **kwargs)


class TestRerunIndividually(TestFlowsException):
    """Repeat tests individually."""

    def __init__(self, tests, *args, **kwargs):
        self.tests = tests
        super(TestRerunIndividually, self).__init__(*args, **kwargs)


class TestFlowsError(TestFlowsException):
    """Base error exception class."""

    pass


class RequirementError(TestFlowsError):
    """Requirement error."""

    pass


class SpecificationError(TestFlowsError):
    """Specification error."""

    pass


class DescriptionError(TestFlowsError):
    """Description error."""

    pass


class ArgumentError(TestFlowsError):
    """Argument error."""

    pass


class TerminatedError(TestFlowsError):
    """Terminated error."""

    pass


class ExecutorError(TestFlowsError):
    """Executor error."""

    pass


class ExecutorWorkerError(ExecutorError):
    """Executor worker error."""

    pass
